@@ -1,17 +1,4 @@
-# Verilog::Getopt.pm -- Verilog command line parsing
-# $Id: Getopt.pm 54310 2008-05-07 18:22:37Z wsnyder $
-# Author: Wilson Snyder <wsnyder@wsnyder.org>
-######################################################################
-#
-# Copyright 2000-2008 by Wilson Snyder.  This program is free software;
-# you can redistribute it and/or modify it under the terms of either the GNU
-# General Public License or the Perl Artistic License.
-# 
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-# 
+# See copyright, etc in below POD section.
 ######################################################################
 
 package Verilog::Getopt;
@@ -29,7 +16,7 @@ use Cwd;
 ######################################################################
 #### Configuration Section
 
-$VERSION = '3.035';
+$VERSION = '3.454';
 
 # Basenames we should ignore when recursing directories,
 # Because they contain large files of no relevance
@@ -58,9 +45,11 @@ sub new {
 		library => [ ],
 		gcc_style => 1,
 		vcs_style => 1,
+		filename_expansion => 0,
 		fileline => 'Command_Line',
 		unparsed => [],
 		define_warnings => 1,
+		depend_files => {},
 		@_
 		};
     bless $self, $class;
@@ -70,11 +59,24 @@ sub new {
 #######################################################################
 # Option parsing
 
+sub _filedir {
+    my $self = shift;
+    my $path = shift;
+    $path =~ s![/\\][^/\\]*$!!   # ~~== my @dirs = File::Spec->splitdir( $path );
+	or $path = ".";
+    return "." if $path eq '';
+    return $path
+}
+
 sub parameter_file {
     my $self = shift;
     my $filename = shift;
+    my $relative = shift;
 
     print "*parameter_file $filename\n" if $Debug;
+    my $optdir = ".";
+    if ($relative) { $optdir = $self->_filedir($filename); }
+
     my $fh = IO::File->new("<$filename") or die "%Error: ".$self->fileline().": $! $filename\n";
     my $hold_fileline = $self->fileline();
     while (my $line = $fh->getline()) {
@@ -83,7 +85,7 @@ sub parameter_file {
 	next if $line =~ /^\s*$/;
 	$self->fileline ("$filename:$.");
 	my @p = (split /\s+/,"$line ");
-	$self->_parameter_parse(@p);
+	$self->_parameter_parse($optdir, @p);
     }
     $fh->close();
     $self->fileline($hold_fileline);
@@ -94,19 +96,22 @@ sub parameter {
     # Parse VCS like parameters, and perform standard setup based on it
     # Return list of leftover parameters
     @{$self->{unparsed}} = ();
-    $self->_parameter_parse(@_);
+    $self->_parameter_parse('.', @_);
     return @{$self->{unparsed}};
 }
 
 sub _parameter_parse {
     my $self = shift;
+    my $optdir = shift;
     # Internal: Parse list of VCS like parameters, and perform standard setup based on it
-    foreach my $param (@_) {
+    foreach my $oparam (@_) {
+	my $param = "$oparam"; # Must quote to convert Getopt to string, bug298
 	next if ($param =~ /^\s*$/);
 	print " parameter($param)\n" if $Debug;
 
 	### GCC & VCS style
-	if ($param eq '-f') {
+	if ($param eq '-F'
+	    || $param eq '-f') {
 	    $self->{_parameter_next} = $param;
 	}
 
@@ -122,11 +127,13 @@ sub _parameter_parse {
 	    }
 	}
 	elsif ($param =~ /^\+incdir\+(.*)$/ && $self->{vcs_style}) {
-	    $self->incdir($1);
+	    $self->incdir($self->_parse_file_arg($optdir, $1));
 	}
-	elsif (($param =~ /^\+define\+([^+=]*)[+=](.*)$/
-		|| $param =~ /^\+define\+(.*?)()$/) && $self->{vcs_style}) {
-	    $self->define ($1, $2);
+	elsif ($param =~ /^\+define\+(.*)$/ && $self->{vcs_style}) {
+	    foreach my $tok (split("\\+", $1)) {
+		my ($a, $b) = $tok =~ m/^([^=]*)=?(.*)$/;
+		$self->define($a,$b,undef,1);
+	    }
 	}
 	# Ignored
 	elsif ($param =~ /^\+librescan$/ && $self->{vcs_style}) {
@@ -135,39 +142,68 @@ sub _parameter_parse {
 	### GCC style
 	elsif (($param =~ /^-D([^=]*)=(.*)$/
 		|| $param =~ /^-D([^=]*)()$/) && $self->{gcc_style}) {
-	    $self->define($1,$2);
+	    $self->define($1,$2,undef,1);
 	}
 	elsif (($param =~ /^-U([^=]*)$/) && $self->{gcc_style}) {
 	    $self->undef($1);
 	}
 	elsif ($param =~ /^-I(.*)$/ && $self->{gcc_style}) {
-	    $self->incdir($1);
+	    $self->incdir($self->_parse_file_arg($optdir, $1));
 	}
 
 	# Second parameters
 	elsif ($self->{_parameter_next}) {
 	    my $pn = $self->{_parameter_next};
 	    $self->{_parameter_next} = undef;
-	    if ($pn eq '-f') {
-		$self->parameter_file ($param);
+	    if ($pn eq '-F') {
+		$self->parameter_file ($self->_parse_file_arg($optdir,$param), 1);
+	    }
+	    elsif ($pn eq '-f') {
+		$self->parameter_file ($self->_parse_file_arg($optdir,$param), 0);
 	    }
 	    elsif ($pn eq '-v') {
-		$self->library ($param);
+		$self->library ($self->_parse_file_arg($optdir,$param));
 	    }
 	    elsif ($pn eq '-y') {
-		$self->module_dir ($param);
+		$self->module_dir ($self->_parse_file_arg($optdir,$param));
 	    }
 	    else {
 		die "%Error: ".$self->fileline().": Bad internal next param ".$pn;
 	    }
 	}
 
-	else { # Unknown
-	    push @{$self->{unparsed}}, $param;
+	else { # Unknown.
+	    if ($self->{filename_expansion}
+		&& $param !~ /^-.*$/ # Presume not a file
+		&& $optdir ne '.') {
+		# If it is a filename, we should ensure it is
+		# relative to $optdir. We assume anything without a leading '-'
+		# is a file, bug 444.
+		my $fn = $self->_parse_file_arg($optdir,$param);
+		if (-e $fn) {
+		    push @{$self->{unparsed}}, "$fn";
+		} else {
+		    push @{$self->{unparsed}}, "$param";
+		}
+	    } else {
+		push @{$self->{unparsed}}, "$param";
+	    }
 	}
     }
 }
 
+sub _parse_file_arg {
+    my $self = shift;
+    my $optdir = shift;
+    my $relfilename = shift;
+    # Parse filename on option line, expanding relative paths in -F's
+    my $filename = $self->file_substitute($relfilename);
+    if ($optdir ne "." && ! File::Spec->file_name_is_absolute($filename)) {
+	$filename = File::Spec->catfile($optdir,$filename);
+    }
+    return $filename;
+}
+
 #######################################################################
 # Accessors
 
@@ -181,7 +217,7 @@ sub incdir {
     if (@_) {
 	my $token = shift;
 	print "incdir $token\n" if $Debug;
-	if (ref $token) {
+	if (ref($token) && ref($token) eq 'ARRAY') {
 	    @{$self->{incdir}} = @{$token};
 	} else {
 	    push @{$self->{incdir}}, $self->file_abs($token);
@@ -195,7 +231,7 @@ sub libext {
     if (@_) {
 	my $token = shift;
 	print "libext $token\n" if $Debug;
-	if (ref $token) {
+	if (ref($token) && ref($token) eq 'ARRAY') {
 	    @{$self->{libext}} = @{$token};
 	} else {
 	    push @{$self->{libext}}, $token;
@@ -209,7 +245,7 @@ sub library {
     if (@_) {
 	my $token = shift;
 	print "library $token\n" if $Debug;
-	if (ref $token) {
+	if (ref($token) && ref($token) eq 'ARRAY') {
 	    @{$self->{library}} = @{$token};
 	} else {
 	    push @{$self->{library}}, $self->file_abs($token);
@@ -222,7 +258,7 @@ sub module_dir {
     if (@_) {
 	my $token = shift;
 	print "module_dir $token\n" if $Debug;
-	if (ref $token) {
+	if (ref($token) && ref($token) eq 'ARRAY') {
 	    @{$self->{module_dir}} = @{$token};
 	} else {
 	    push @{$self->{module_dir}}, $self->file_abs($token);
@@ -234,7 +270,8 @@ sub module_dir {
 sub depend_files {
     my $self = shift;
     if (@_) {
-	if (ref $_[0]) {
+	#@_ may be Getopt::Long::Parameters which aren't arrays, will stringify
+	if (ref($_[0]) && ref($_[0]) eq 'ARRAY') {
 	    $self->{depend_files} = {};
 	    foreach my $fn (@{$_[0]}) {
 		$self->{depend_files}{$fn} = 1;
@@ -255,7 +292,7 @@ sub get_parameters {
     my %args = (gcc_stlyle => $self->{gcc_style},);
     # Defines
     my @params = ();
-    foreach my $def (sort (keys %{$self->{defines}})) {
+    foreach my $def ($self->define_names_sorted) {
 	my $defvalue = $self->defvalue($def);
 	$defvalue = "=".($defvalue||"") if (defined $defvalue && $defvalue ne "");
 	if ($args{gcc_style}) {
@@ -302,7 +339,7 @@ sub write_parameters_file {
 # Utility functions
 
 sub remove_duplicates {
-    my $self = shift if ref $_[0];
+    my $self = ref $_[0] && shift;
     # return list in same order, with any duplicates removed
     my @rtn;
     my %hit;
@@ -334,6 +371,20 @@ sub file_abs {
     return $filename;
 }
 
+sub file_substitute {
+    my $self = shift;
+    my $filename = shift;
+    my $out = $filename;
+    while ($filename =~ /\$([A-Za-z_0-9]+)\b/g) {
+	my $var = $1;
+	if (defined $ENV{$var}) {
+	    $out =~ s/\$$var\b/$ENV{$var}/g;
+	}
+    }
+    $out =~ s!^~!$ENV{HOME}/!;
+    return $out;
+}
+
 sub file_path_cache_flush {
     my $self = shift;
     # Clear out a file_path cache, needed if the incdir/module_dirs change
@@ -354,6 +405,14 @@ sub file_path {
 	$self->depend_files($filename);
 	return $filename;
     }
+    # Try expanding environment
+    $filename = $self->file_substitute($filename);
+    if (-r $filename && !-d $filename) {
+	$self->{_file_path_cache}{$filename} = $filename;
+	$self->depend_files($filename);
+	return $filename;
+    }
+
     # What paths to use?
     my @dirlist;
     if ($lookup_type eq 'module') {
@@ -361,8 +420,12 @@ sub file_path {
     } elsif ($lookup_type eq 'include') {
 	@dirlist = $self->incdir();
     } else {  # all
+	# Might be more obvious if -y had priority, but we'll remain back compatible
 	@dirlist = ($self->incdir(), $self->module_dir());
     }
+    # Expand any envvars in incdir/moduledir
+    @dirlist = map {$self->file_substitute($_)} @dirlist;
+
     # Check each search path
     # We use both the incdir and moduledir.  This isn't strictly correct,
     # but it's fairly silly to have to specify both all of the time.
@@ -381,7 +444,7 @@ sub file_path {
 	    }
 	}
     }
-    
+
     return $filename;	# Let whoever needs it discover it doesn't exist
 }
 
@@ -415,14 +478,30 @@ sub map_directories {
 #######################################################################
 # Getopt functions
 
+sub define_names_sorted {
+    my $self = shift;
+    return (sort (keys %{$self->{defines}}));
+}
+
+sub defcmdline {
+    my $self = shift;
+    my $token = shift;
+    my $val = $self->{defines}{$token};
+    if (ref $val) {
+	return $val->[2];
+    } else {
+	return undef;
+    }
+}
+
 sub defparams {
     my $self = shift;
     my $token = shift;
     my $val = $self->{defines}{$token};
     if (!defined $val) {
 	return undef;
-    } elsif (ref $val) {
-	return $val->[1];  # Has parameters, return param list
+    } elsif (ref $val && defined $val->[1]) {
+	return $val->[1];  # Has parameters hash, return param list or undef
     } else {
 	return 0;
     }
@@ -453,8 +532,9 @@ sub define {
     if (@_) {
 	my $token = shift;
 	my $value = shift;
-	my $params = shift||"";
-	print "Define $token $params= $value\n" if $Debug;
+	my $params = shift;
+	my $cmdline = shift;
+	print "Define $token ".($params||'')."= $value\n" if $Debug;
 	my $oldval = $self->{defines}{$token};
 	my $oldparams;
 	if (ref $oldval eq 'ARRAY') {
@@ -462,12 +542,16 @@ sub define {
 	}
 	if (defined $oldval
 	    && (($oldval ne $value)
-		|| (defined $oldparams && $oldparams ne $params))
+		|| (($oldparams||'') ne ($params||'')))
 	    && $self->{define_warnings}) {
-	    warn "%Warning: ".$self->fileline().": Redefining `$token\n";
+	    warn "%Warning: ".$self->fileline().": Redefining `$token"
+		# Don't make errors too long or have strange chars
+		.((length($oldval)<40 && $oldval =~ /^[^\n\r\f]$/
+		   && length($value)<40 && $value =~ /^[^\n\r\f]$/)
+		  ? "to '$value', was '$oldval'\n" : "\n");
 	}
-	if ($params) {
-	    $self->{defines}{$token} = [$value, $params];
+	if ($params || $cmdline) {
+	    $self->{defines}{$token} = [$value, $params, $cmdline];
 	} else {
 	    $self->{defines}{$token} = $value;
 	}
@@ -483,6 +567,15 @@ sub undef {
     delete $self->{defines}{$token};
 }
 
+sub undefineall {
+    my $self = shift;
+    foreach my $def (keys %{$self->{defines}}) {
+	if (!$self->defcmdline($def)) {
+	    delete $self->{defines}{$def};
+	}
+    }
+}
+
 sub remove_defines {
     my $self = shift;
     my $sym = shift;
@@ -523,20 +616,44 @@ Verilog::Getopt - Get Verilog command line options
 Verilog::Getopt provides standardized handling of options similar to
 Verilog/VCS and cc/GCC.
 
+=head1 OPTIONS
+
+The new() constructor accepts the following options:
+
+=over 4
+
+=item filename_expansion=>1
+
+Enable converting filenames to relative filenames when possible.  This
+option is needed when the -F option will be used.  If flags are passed
+through Getopt which should otherwise not be expanded (e.g. "--out
+myfile.v") having this option set may undesirably expand myfile.v to an
+absolute filename.
+
+=item gcc_style=>0
+
+Disable parsing of GCC-like parameters.
+
+=item vcs_style=>0
+
+Disable parsing of VCS-like parameters.
+
+=back
+
+=head1 METHODS
+
 =over 4
 
 =item $opt = Verilog::Getopt->new ( I<opts> )
 
-Create a new Getopt.  If gcc_style=>0 is passed as a parameter, parsing of
-GCC-like parameters is disabled.  If vcs_style=>0 is passed as a parameter,
-parsing of VCS-like parameters is disabled.
+Create a new Getopt.  See OPTIONS above.
 
-=item $self->file_path ( I<filename>, [I<lookup-type>] )
+=item $self->file_path ( I<filename>, [I<lookup_type>] )
 
 Returns a new path to the filename, using the library directories and
-search paths to resolve the file.  Optional lookup-type is
-'module','include', or 'all', to use only module_dirs, incdirs, or both for
-the lookup.
+search paths to resolve the file.  Optional lookup_type is 'module',
+'include', or 'all', to use only module_dirs, incdirs, or both for the
+lookup.
 
 =item $self->get_parameters ( )
 
@@ -547,16 +664,18 @@ for downstream programs that also use Verilog::Getopt.
 =item $self->parameter ( \@params )
 
 Parses any recognized parameters in the referenced array, removing the
-standard parameters and returning a array with all unparsed parameters.
+standard parameters from any previous parameters() call, and returning a
+array with all unparsed parameters.
 
 The below list shows the VCS-like parameters that are supported, and the
 functions that are called:
 
     +libext+I<ext>+I<ext>...	libext (I<ext>)
     +incdir+I<dir>		incdir (I<dir>)
-    +define+I<var>[+=]I<value>	define (I<var>,I<value>)
+    +define+I<var>=I<value>	define (I<var>,I<value>)
     +define+I<var>		define (I<var>,undef)
     +librescan		Ignored
+    -F I<file>		Parse parameters in file relatively
     -f I<file>		Parse parameters in file
     -v I<file>		library (I<file>)
     -y I<dir>		module_dir (I<dir>)
@@ -569,6 +688,7 @@ functions that are called:
     -DI<var>		define (I<var>,undef)
     -UI<var>		undefine (I<var>)
     -II<dir>		incdir (I<dir>)
+    -F I<file>		Parse parameters in file relatively
     -f I<file>		Parse parameters in file
     all others		Put in returned list
 
@@ -588,6 +708,14 @@ This method is called when a define is recognized.  The default behavior
 loads a hash that is used to fulfill define references.  This function may
 also be called outside parsing to predefine values.
 
+An optional third argument specifies parameters to the define, and a fourth
+argument if true indicates the define was set on the command line and
+should not be removed by `undefineall.
+
+=item $self->define_names_sorted
+
+Return sorted list of all define names that currently exist.
+
 =item $self->defparams ( $token )
 
 This method returns the parameter list of the define.  This will be defined,
@@ -617,6 +745,11 @@ Using the incdir and libext lists, convert the specified module or filename
 Return true if the filename is one that generally should be ignored when
 recursing directories, such as for example, ".", "CVS", and ".svn".
 
+=item $self->file_substitute ( $filename )
+
+Removes existing environment variables from the provided filename.  Any
+undefined variables are not substituted nor cause errors.
+
 =item $self->incdir ()
 
 Returns reference to list of include directories.  With argument, adds that
@@ -649,6 +782,10 @@ Return string with any definitions in the token removed.
 Deletes a hash element that is used to fulfill define references.  This
 function may also be called outside parsing to erase a predefined value.
 
+=item $self->undefineall ()
+
+Deletes all non-command line definitions, for implementing `undefineall.
+
 =back
 
 =head1 DISTRIBUTION
@@ -657,9 +794,9 @@ Verilog-Perl is part of the L<http://www.veripool.org/> free Verilog EDA
 software tool suite.  The latest version is available from CPAN and from
 L<http://www.veripool.org/verilog-perl>.
 
-Copyright 2000-2008 by Wilson Snyder.  This package is free software; you
+Copyright 2000-2018 by Wilson Snyder.  This package is free software; you
 can redistribute it and/or modify it under the terms of either the GNU
-Lesser General Public License or the Perl Artistic License.
+Lesser General Public License Version 3 or the Perl Artistic License Version 2.0.
 
 =head1 AUTHORS
 
