@@ -1,17 +1,4 @@
-# Verilog::Language.pm -- Verilog language keywords, etc
-# $Id: Language.pm 54310 2008-05-07 18:22:37Z wsnyder $
-# Author: Wilson Snyder <wsnyder@wsnyder.org>
-######################################################################
-#
-# Copyright 2000-2008 by Wilson Snyder.  This program is free software;
-# you can redistribute it and/or modify it under the terms of either the GNU
-# General Public License or the Perl Artistic License.
-# 
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-# 
+# See copyright, etc in below POD section.
 ######################################################################
 
 =pod
@@ -39,53 +26,6 @@ Verilog::Language provides general utilities for using the Verilog
 Language, such as parsing numbers or determining what keywords exist.
 General functions will be added as needed.
 
-=head1 WHICH PACKAGE
-
-If you are starting a new application which needs to parse the Verilog
-language you have several tools available to you.  Which you pick depends
-on how low level and complete the information you need is.
-
-=over 4
-
-=item Verilog::Preproc
-
-Verilog::Preproc is useful when you need only text out, or a list of
-defines, etc.  It can preprocess a file, or be used to provide the Verilog
-macro language on top of synthesis scripts.  It understands the full
-SystemVerilog 2005 preprocessor syntax.
-
-=item Verilog::Parser
-
-Verilog::Parser is useful when you need to tokenize or write source filters
-(where you need everything including whitespace).  It can take raw files,
-or preprocessed input.  It understands all SystemVerilog 2005 keywords.
-
-=item Verilog::SigParser
-
-Verilog::SigParser is useful when you need a list of modules, signals,
-ports, functions, etc.  It requires a preprocessed file, and can parse most
-Verilog 2005 files, but only provides callbacks on certain interesting
-things.
-
-=item Verilog::Netlist
-
-Verilog::Netlist is useful for when you need the hierarchy, and a list of
-signals per module, pins per cell, etc.  It builds upon the output of
-Verilog::SigParser, so requires preprocessed files.
-
-This is probably the most popular choice.
-
-=item VPI
-
-Using the VPI is the best way to access the behavior of the design.  It is
-not part of this package as it requires a compliant simulator and C++ code
-to call the VPI, and understands as much of the language as the simulator
-supports.  This allows writing lint checks and full knowledge of all parts
-of the code, but generally requires the most work (short of writing a
-parser from scratch.)
-
-=back
-
 =head1 FUNCTIONS
 
 =over 4
@@ -94,7 +34,8 @@ parser from scratch.)
 
 Return true if the given symbol string is a Verilog reserved keyword.
 Value indicates the language standard as per the `begin_keywords macro,
-'1364-1995', '1364-2001', '1364-2005', or '1800-2005'.
+'1364-1995', '1364-2001', '1364-2005', '1800-2005', '1800-2009',
+'1800-2012', '1800-2017' or 'VAMS'.
 
 =item Verilog::Language::is_compdirect ($symbol_string)
 
@@ -105,12 +46,22 @@ Return true if the given symbol string is a Verilog compiler directive.
 Return true if the given symbol is a built in gate primitive; for example
 "buf", "xor", etc.
 
+=item Verilog::Language::language_keywords ($year)
+
+Returns a hash for keywords for given language standard year, where the
+value of the hash is the standard in which it was defined.
+
 =item Verilog::Language::language_standard ($year)
 
 Sets the language standard to indicate what are keywords.  If undef, all
 standards apply.  The year is indicates the language standard as per the
-`begin_keywords macro, '1364-1995', '1364-2001', '1364-2005', or
-'1800-2005'.
+`begin_keywords macro, '1364-1995', '1364-2001', '1364-2005', '1800-2005'
+'1800-2009', '1800-2012' or '1800-2017'.
+
+=item Verilog::Language::language_maximum
+
+Returns the greatest language currently standardized, presently
+'1800-2017'.
 
 =item Verilog::Language::number_bigint ($number_string)
 
@@ -168,9 +119,9 @@ Verilog-Perl is part of the L<http://www.veripool.org/> free Verilog EDA
 software tool suite.  The latest version is available from CPAN and from
 L<http://www.veripool.org/verilog-perl>.
 
-Copyright 2000-2008 by Wilson Snyder.  This package is free software; you
+Copyright 2000-2018 by Wilson Snyder.  This package is free software; you
 can redistribute it and/or modify it under the terms of either the GNU
-Lesser General Public License or the Perl Artistic License.
+Lesser General Public License Version 3 or the Perl Artistic License Version 2.0.
 
 =head1 AUTHORS
 
@@ -180,8 +131,8 @@ Wilson Snyder <wsnyder@wsnyder.org>
 
 L<Verilog-Perl>,
 L<Verilog::EditFiles>
-L<Verilog::Parser>, 
-L<Verilog::ParseSig>, 
+L<Verilog::Parser>,
+L<Verilog::ParseSig>,
 L<Verilog::Getopt>
 
 And the L<http://www.veripool.org/verilog-mode>Verilog-Mode package for Emacs.
@@ -200,7 +151,7 @@ use Carp;
 ######################################################################
 #### Configuration Section
 
-$VERSION = '3.035';
+$VERSION = '3.454';
 
 ######################################################################
 #### Internal Variables
@@ -254,21 +205,94 @@ foreach my $kwd (qw(
 		    wait_order wildcard with within
 		    )) { $Keywords{'1800-2005'}{$kwd} = '1800-2005'; }
 
+foreach my $kwd (qw(
+		    accept_on checker endchecker eventually global implies
+		    let nexttime reject_on restrict s_always s_eventually
+		    s_nexttime s_until s_until_with strong sync_accept_on
+		    sync_reject_on unique0 until until_with untyped weak
+		    )) { $Keywords{'1800-2009'}{$kwd} = '1800-2009'; }
+
+foreach my $kwd (qw(
+		    implements nettype interconnect soft
+		    )) { $Keywords{'1800-2012'}{$kwd} = '1800-2012'; }
+
+foreach my $kwd (qw(
+		    )) { $Keywords{'1800-2017'}{$kwd} = '1800-2017'; }
+
+foreach my $kwd (qw(
+		    above abs absdelay abstol ac_stim access acos acosh
+		    aliasparam analog analysis asin asinh assert atan atan2
+		    atanh branch ceil connect connectmodule connectrules
+		    continuous cos cosh cross ddt ddt_nature ddx discipline
+		    discrete domain driver_update endconnectrules
+		    enddiscipline endnature endparamset exclude exp
+		    final_step flicker_noise floor flow from ground hypot
+		    idt idt_nature idtmod inf initial_step laplace_nd
+		    laplace_np laplace_zd laplace_zp last_crossing limexp
+		    ln log max merged min nature net_resolution noise_table
+		    paramset potential pow resolveto sin sinh slew split
+		    sqrt string tan tanh timer transition units white_noise
+		    wreal zi_nd zi_np zi_zd zi_zp
+		    )) { $Keywords{'VAMS'}{$kwd} = 'VAMS'; }
+
 foreach my $kwd (
-		 "`celldefine", "`default_nettype", "`define", "`else",
-		 "`endcelldefine", "`endif", "`ifdef", "`include",
-		 "`nounconnected_drive", "`resetall", "`timescale",
-		 "`unconnected_drive", "`undef",
-		 # Commercial Extensions
-		 "`protected", "`endprotected",
-		 ) { $Keywords{$kwd}{'1364-1995'} = $Compdirect{$kwd} = '1364-1995'; }
+    # Speced
+    "`celldefine",
+    "`define",			# Preprocessor
+    "`else",			# Preprocessor
+    "`endcelldefine",
+    "`endif",			# Preprocessor
+    "`ifdef",			# Preprocessor
+    "`include",			# Preprocessor
+    "`nounconnected_drive",
+    "`resetall",
+    "`timescale",
+    "`unconnected_drive",
+    "`undef",			# Preprocessor
+    "`undefineall",		# Preprocessor
+
+    # Commercial Extensions
+    "`accelerate",		# Verilog-XL compatibility
+    "`autoexpand_vectornets",	# Verilog-XL compatibility
+    "`default_decay_time",	# Verilog spec - delays only
+    "`default_trireg_strength",	# Verilog spec
+    "`delay_mode_distributed",	# Verilog spec - delays only
+    "`delay_mode_path",		# Verilog spec - delays only
+    "`delay_mode_unit",		# Verilog spec - delays only
+    "`delay_mode_zero",		# Verilog spec - delays only
+    "`disable_portfaults",	# Verilog-XL compatibility
+    "`enable_portfaults",	# Verilog-XL compatibility
+    "`endprotect",		# Many tools - pre encryption
+    "`endprotected",		# Many tools - post encryption
+    "`expand_vectornets",	# Verilog-XL compatibility
+    "`noaccelerate",		# Verilog-XL compatibility
+    "`noexpand_vectornets",	# Verilog-XL compatibility
+    "`noremove_gatenames",	# Verilog-XL compatibility
+    "`noremove_netnames",	# Verilog-XL compatibility
+    "`nosuppress_faults",	# Verilog-XL compatibility
+    "`nounconnected_drive",	# Verilog-XL compatibility
+    "`portcoerce",		# Verilog-XL compatibility
+    "`protect",			# Many tools - pre encryption
+    "`protected",		# Many tools - post encryption
+    "`remove_gatenames",	# Verilog-XL compatibility
+    "`remove_netnames",		# Verilog-XL compatibility
+    "`suppress_faults",		# Verilog-XL compatibility
+    ) { $Keywords{$kwd}{'1364-1995'} = $Compdirect{$kwd} = '1364-1995'; }
 
 foreach my $kwd (
 		 "`default_nettype", "`elsif", "`undef", "`ifndef",
 		 "`file", "`line",
 		 ) { $Keywords{$kwd}{'1364-2001'} = $Compdirect{$kwd} = '1364-2001'; }
 
-language_standard ('1800-2005');  # Default standard
+foreach my $kwd (
+		 "`pragma",
+		 ) { $Keywords{$kwd}{'1364-2005'} = $Compdirect{$kwd} = '1364-2005'; }
+
+foreach my $kwd (
+		 "`default_discipline", "`default_transition",
+		 ) { $Keywords{$kwd}{'1364-2005'} = $Compdirect{$kwd} = '1364-2005'; }
+
+language_standard (language_maximum());  # Default standard
 
 foreach my $kwd (qw(
 		    and buf bufif0 bufif1 cmos nand nmos nor not notif0
@@ -279,37 +303,71 @@ foreach my $kwd (qw(
 ######################################################################
 #### Keyword utilities
 
+sub language_maximum {
+    return "1800-2017";
+}
+
+sub _language_kwd_hash {
+    my $standard = shift;
+    my @subsets;
+    if ($standard eq '1995' || $standard eq '1364-1995') {
+	$Standard = '1364-1995';
+	@subsets = ('1364-1995');
+    } elsif ($standard eq '2001' || $standard eq '1364-2001' || $standard eq '1364-2001-noconfig') {
+	$Standard = '1364-2001';
+	@subsets = ('1364-2001',
+		    '1364-1995');
+    } elsif ($standard eq '1364-2005') {
+	$Standard = '1364-2005';
+	@subsets = ('1364-2005',
+		    '1364-2001', '1364-1995');
+    } elsif ($standard eq 'sv31' || $standard eq '1800-2005') {
+	$Standard = '1800-2005';
+	@subsets = ('1800-2005',
+		    '1364-2005', '1364-2001', '1364-1995');
+    } elsif ($standard eq '1800-2009') {
+	$Standard = '1800-2009';
+	@subsets = ('1800-2009', '1800-2005',
+		    '1364-2005', '1364-2001', '1364-1995');
+    } elsif ($standard eq '1800-2012') {
+	$Standard = '1800-2012';
+	@subsets = ('1800-2012', '1800-2009', '1800-2005',
+		    '1364-2005', '1364-2001', '1364-1995');
+    } elsif ($standard eq 'latest' || $standard eq '1800-2017') {
+	$Standard = '1800-2017';
+	@subsets = ('1800-2017', '1800-2012', '1800-2009', '1800-2005',
+		    '1364-2005', '1364-2001', '1364-1995');
+    } elsif ($standard =~ /^V?AMS/) {
+	$Standard = 'VAMS';
+	@subsets = ('VAMS',
+		    '1364-2005', '1364-2001', '1364-1995');
+    } else {
+	croak "%Error: Verilog::Language::language_standard passed bad value: $standard,";
+    }
+    # Update keyword list to present language
+    # (We presume the language_standard rarely changes, so it's faster to compute the list.)
+    my %keywords = ();
+    foreach my $ss (@subsets) {
+	foreach my $kwd (%{$Keywords{$ss}}) {
+	    $keywords{$kwd} = $ss;
+	}
+    }
+    return %keywords;
+}
+
 sub language_standard {
     my $standard = shift;
     if (defined $standard) {
-	my @subsets;
-	if ($standard eq '1995' || $standard eq '1364-1995') {
-	    $Standard = '1364-1995';
-	    @subsets = ('1364-1995');
-	} elsif ($standard eq '2001' || $standard eq '1364-2001' || $standard eq '1364-2001-noconfig') {
-	    $Standard = '1364-2001';
-	    @subsets = ('1364-2001', '1364-1995');
-	} elsif ($standard eq '1364-2005') {
-	    $Standard = '1364-2005';
-	    @subsets = ('1364-2005', '1364-2001', '1364-1995');
-	} elsif ($standard eq 'sv31' || $standard eq '1800-2005') {
-	    $Standard = '1800-2005';
-	    @subsets = ('1800-2005', '1364-2005', '1364-2001', '1364-1995');
-	} else {
-	    croak "%Error: Verilog::Language::language_standard passed bad value: $standard,";
-	}
-	# Update keyword list to present language
-	# (We presume the language_standard rarely changes, so it's faster to compute the list.)
-	%Keyword = ();
-	foreach my $ss (@subsets) {
-	    foreach my $kwd (%{$Keywords{$ss}}) {
-		$Keyword{$kwd} = $ss;
-	    }
-	}
+	%Keyword = _language_kwd_hash($standard);
     }
     return $Standard;
 }
 
+sub language_keywords {
+    my $standard = shift || $Standard;
+    return _language_kwd_hash($standard);
+}
+
 sub is_keyword {
     my $symbol = shift;
     return ($Keyword{$symbol});
@@ -386,15 +444,14 @@ sub number_value {
     }
     elsif ($number =~ /\'s?b([0-1]+)$/i) {
 	my $val = 0;
-	my $bit;
 	$number = $1;
-	foreach $bit (split(//, $number)) {
+	foreach my $bit (split(//, $number)) {
 	    $val = ($val<<1) | ($bit=='1'?1:0);
 	}
 	return ($val);
     }
     elsif ($number =~ /\'s?d?([0-9]+)$/i
-	   || $number =~ /^([0-9]+)$/i) {
+	   || $number =~ /^(-?[0-9]+)$/i) {
 	return ($1);
     }
     return undef;
@@ -493,7 +550,7 @@ sub split_bus {
 	    if (/^\[/) {
 		$inbra = 1;
 		$pretext[$brnum] .= $_;
-	    } 
+	    }
 	    if (!$inbra) {
 		# Not in bracket, just remember text
 		$pretext[$brnum] .= $_;
@@ -599,7 +656,7 @@ sub split_bus_nocomma {
 	    if (/^\[/) {
 		$inbra = 1;
 		$pretext[$brnum] .= $_;
-	    } 
+	    }
 	    if (!$inbra) {
 		# Not in bracket, just remember text
 		$pretext[$brnum] .= $_;
